from dataclasses import dataclass, field
from typing import List, Tuple, Any

from marshy import get_default_context
from marshy.factory.impl_marshaller_factory import ImplMarshallerFactory
from sqlalchemy import Table

from persisty.impl.sqlalchemy.search_filter.search_filter_converter_abc import (
    SearchFilterConverterABC,
)
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta


def get_converters() -> List[SearchFilterConverterABC]:
    marshaller_context = get_default_context()
    factories = marshaller_context.get_factories()
    for factory in factories:
        if isinstance(factory, ImplMarshallerFactory):
            if factory.base == SearchFilterConverterABC:
                converters = [impl() for impl in factory.impls]
                converters.sort(key=lambda i: i.priority)
                return converters


@dataclass
class SearchFilterConverterContext:
    converters: List[SearchFilterConverterABC] = field(default_factory=get_converters)

    def convert(
        self, search_filter: SearchFilterABC, table: Table, store_meta: StoreMeta
    ) -> Tuple[Any, bool]:
        for converter in self.converters:
            result = converter.convert(search_filter, table, store_meta, self)
            if result:
                return result
        return None, False



from dataclasses import dataclass

from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC


@dataclass
class IntSequenceGenerator(AttrValueGeneratorABC):
    """
    Sequence Id Generator. Note: There could be a security concern with making this optional for create operations - for
    filtered storage, it could open a way for attackers to check if an id exists. (By trying to create it)
    Typically sql based keys will generate this value in the database rather than relying on a client.
    """

    always: bool = True
    value: int = 1
    step: int = 1

    def transform(self, value, item):
        if self.always or not value:
            value = self.value
            self.value += self.step
        return value



from __future__ import annotations

from typing import Tuple, Optional, Any, TYPE_CHECKING
from uuid import uuid4

from servey.util.singleton_abc import SingletonABC

from persisty.search_filter.search_filter_abc import SearchFilterABC, T

if TYPE_CHECKING:
    from persisty.attr.attr import Attr


class ExcludeAll(SearchFilterABC[T], SingletonABC):
    def lock_attrs(self, attrs: Tuple[Attr, ...]) -> ExcludeAll:
        return self

    # pylint: disable=W0613
    def match(self, item: T, attrs: Tuple[Attr, ...]) -> bool:
        return False

    def build_filter_expression(
        self, attrs: Tuple[Attr, ...]
    ) -> Tuple[Optional[Any], bool]:
        """This should be caught as it means no query should run"""
        from boto3.dynamodb.conditions import Attr

        return Attr(str(uuid4())).eq(1), True


EXCLUDE_ALL = ExcludeAll()



from dataclasses import dataclass, field
from typing import Optional

from persisty.impl.sqlalchemy.sqlalchemy_context import SqlalchemyContext
from persisty.impl.sqlalchemy.sqlalchemy_context_factory_abc import (
    get_default_context,
)
from persisty.impl.sqlalchemy.sqlalchemy_table_store import SqlalchemyTableStore
from persisty.store.restrict_access_store import restrict_access_store
from persisty.store.schema_validating_store import SchemaValidatingStore
from persisty.store.store_abc import StoreABC
from persisty.store_meta import StoreMeta
from persisty.trigger.wrapper import triggered_store


@dataclass
class SqlalchemyTableStoreFactory:
    store_meta: StoreMeta
    context: SqlalchemyContext = field(default_factory=get_default_context)

    def get_meta(self) -> StoreMeta:
        return self.store_meta

    def create(self) -> Optional[StoreABC]:
        table = self.context.get_table(self.store_meta)
        store = SqlalchemyTableStore(self.store_meta, table, self.context.engine)
        store = SchemaValidatingStore(store)
        store = restrict_access_store(store, self.store_meta.store_access)
        store = triggered_store(store)
        return store



from __future__ import annotations

from typing import Tuple, Optional, Any

from servey.util.singleton_abc import SingletonABC

from persisty.attr.attr import Attr
from persisty.search_filter.search_filter_abc import SearchFilterABC, T


class IncludeAll(SearchFilterABC[T], SingletonABC):
    def lock_attrs(self, attrs: Tuple[Attr, ...]) -> IncludeAll:
        return self

    def match(self, item: T, attrs: Tuple[Attr, ...]) -> bool:
        return True

    def build_filter_expression(
        self, attrs: Tuple[Attr, ...]
    ) -> Tuple[Optional[Any], bool]:
        return None, True


INCLUDE_ALL = IncludeAll()



import os
from dataclasses import dataclass

from servey.servey_aws import is_lambda_env

from persisty.impl.mem.mem_store_factory import MemStoreFactory
from persisty.store.store_abc import StoreABC, T
from persisty.store.wrapper_store_abc import WrapperStoreABC
from persisty.store_meta import StoreMeta


@dataclass
class DefaultStore(WrapperStoreABC[T]):
    meta: StoreMeta

    def get_meta(self) -> StoreMeta:
        return self.meta

    def get_store(self) -> StoreABC:
        store = getattr(self, "_store", None)
        if store:
            return store
        if os.environ.get("PERSISTY_SQL_URN"):
            from persisty.impl.sqlalchemy.sqlalchemy_table_store_factory import (
                SqlalchemyTableStoreFactory,
            )

            factory = SqlalchemyTableStoreFactory(self.meta)
        elif is_lambda_env():
            from persisty.impl.dynamodb.dynamodb_store_factory import (
                DynamodbStoreFactory,
            )

            factory = DynamodbStoreFactory(self.meta)
            factory.derive_from_meta()
        else:
            from persisty.io.seed import get_seed_items

            seed_items = get_seed_items(self.meta)
            key_config = self.meta.key_config
            seed_data = {key_config.to_key_str(i): i for i in seed_items}
            factory = MemStoreFactory(self.meta, seed_data)
        store = factory.create()
        setattr(self, "_store", store)
        return store



from dataclasses import dataclass
from typing import List, Dict, Iterable, Optional

from marshy.types import ExternalItemType

from persisty.attr.attr import Attr
from persisty.impl.dynamodb.dynamodb_key_config import DynamodbKeyConfig
from persisty.index.index_abc import IndexABC
from persisty.key_config.attr_key_config import AttrKeyConfig


@dataclass(frozen=True)
class PartitionSortIndex(IndexABC):
    """
    Index which functions similarly to an AttrIndex, but has an optional sort key (Useful for dynamodb)
    Sql tables treat this as an attr index
    """

    pk: str
    sk: Optional[str] = None

    def to_schema(self):
        schema = [{"AttributeName": self.pk, "KeyType": "HASH"}]
        if self.sk:
            schema.append({"AttributeName": self.sk, "KeyType": "RANGE"})
        return schema

    def to_condition_expression(self, item: ExternalItemType):
        from boto3.dynamodb.conditions import Key, And as DynAnd

        condition_expression = Key(self.pk).eq(item[self.pk])
        if self.sk and item.get(self.sk) is not None:
            sk = Key(self.sk).eq(item[self.sk])
            condition_expression = DynAnd(condition_expression, sk)
        return condition_expression

    def to_dict(self, item: ExternalItemType):
        d = {self.pk: item[self.pk]}
        if self.sk:
            d[self.sk] = item[self.sk]
        return d

    def key_config_from_attrs(self, attrs: Iterable[Attr]):
        pk_attr = next(a for a in attrs if a.name == self.pk)
        pk = AttrKeyConfig(self.pk, pk_attr.attr_type)
        if not self.sk:
            return pk
        sk_attr = next(a for a in attrs if a.name == self.sk)
        sk = AttrKeyConfig(self.sk, sk_attr.attr_type)
        return DynamodbKeyConfig(pk, sk)

    def _to_key_config(self, configs: List[AttrKeyConfig]):
        pk_attr = next(a for a in configs if a.attr_name == self.pk)
        if not self.sk:
            return pk_attr
        sk_attr = next(a for a in configs if a.attr_name == self.sk)
        return DynamodbKeyConfig(pk_attr, sk_attr)


def from_schema(schema: List[Dict]):
    assert len(schema) == 1 or len(schema) == 2
    index = PartitionSortIndex(
        pk=next(a["AttributeName"] for a in schema if a["KeyType"] == "HASH"),
        sk=next((a["AttributeName"] for a in schema if a["KeyType"] == "RANGE"), None),
    )
    return index


ID_INDEX = PartitionSortIndex("id")



from datetime import datetime

from persisty.finder.store_finder_abc import find_stores
from persisty.impl.default_store import DefaultStore
from persisty.impl.sqlalchemy.sqlalchemy_context_factory_abc import get_default_context
from persisty.io.seed import get_seed_data


def get_target_metadata():
    """
    Reference this in the alembic env.py
    """
    for store in find_stores():
        if isinstance(store, DefaultStore):
            store.get_store()  # make sure tables are registered
    target_metadata = get_default_context().meta_data
    return target_metadata


def add_seed_data(op):
    target_metadata = get_target_metadata()
    for store in find_stores():
        meta = store.get_meta()
        table = target_metadata.tables.get(meta.name)
        cols = list(table.columns)
        items = []
        for item in get_seed_data(meta.name):
            filtered_item = {**item}
            for col in cols:
                if col.type.python_type == datetime:
                    value = filtered_item.get(col.name)
                    if value:
                        filtered_item[col.name] = datetime.fromisoformat(value)
            items.append(filtered_item)
        op.bulk_insert(table, items)



from dataclasses import field
from typing import Dict

from marshy.types import ExternalItemType
from servey.action.action import Action
from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
    TriggerHandlerABC,
)
from servey.trigger.trigger_abc import TriggerABC

from persisty.trigger.after_create_trigger import AfterCreateTrigger
from persisty.trigger.after_delete_trigger import AfterDeleteTrigger
from persisty.trigger.after_update_trigger import AfterUpdateTrigger


class ServerlessTriggerHandler(TriggerHandlerABC):
    unmanaged_table_arns_by_store_name: Dict[str, str] = field(default_factory=dict)

    def handle_trigger(
        self, action: Action, trigger: TriggerABC, lambda_definition: ExternalItemType
    ):
        if (
            not isinstance(trigger, AfterCreateTrigger)
            and not isinstance(trigger, AfterUpdateTrigger)
            and not isinstance(trigger, AfterDeleteTrigger)
        ):
            return
        stream = lambda_definition.get("stream")
        if not stream:
            stream = lambda_definition["stream"] = []
        # pylint: disable=E1101
        arn = self.unmanaged_table_arns_by_store_name.get(trigger.store_name)
        if not arn:
            arn = {
                "Fn::GetAtt": [
                    trigger.store_name.title().replace("_", ""),
                    "StreamArn",
                ]
            }
        stream.append(
            {
                "type": "dynamodb",
                "arn": arn,
            }
        )



from dataclasses import dataclass
from enum import Enum
from typing import Type, Optional

from persisty.search_order.search_order import SearchOrder
from persisty.search_order.search_order_attr import SearchOrderAttr
from persisty.store_meta import StoreMeta
from persisty.util import to_camel_case


class SearchOrderFactoryABC:
    """
    A lot of formats rely on search order having a simple flat structure. This facilitates that pattern
    """

    attr: Enum
    desc: bool

    def to_search_order(self) -> SearchOrder:
        attr = self.attr
        if attr:
            return SearchOrder((SearchOrderAttr(self.attr.name, self.desc),))


def search_order_dataclass_for(
    store_meta: StoreMeta,
) -> Optional[Type[SearchOrderFactoryABC]]:
    sortable_attrs = store_meta.get_sortable_attrs_as_enum()
    if not sortable_attrs:
        return
    params = {
        "__annotations__": {
            "attr": Optional[sortable_attrs],
            "desc": bool,
        },
        "__doc__": f"Search Order for {store_meta.name}",
        "attr": None,
        "desc": False,
    }
    name = f"{to_camel_case(store_meta.name)}SearchOrder"
    type_ = dataclass(type(name, (SearchOrderFactoryABC,), params))
    return type_



from datetime import datetime
from typing import Type, Optional, Tuple
from uuid import UUID

from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC
from persisty.attr.generator.int_sequence_generator import IntSequenceGenerator
from persisty.attr.generator.str_sequence_generator import StrSequenceGenerator
from persisty.attr.generator.timestamp_generator import TimestampGenerator
from persisty.attr.generator.uuid_generator import UuidGenerator


def get_default_generator_for_create(
    name: str, type_: Type
) -> Tuple[bool, Optional[AttrValueGeneratorABC]]:
    if name == "id":
        if type_ == UUID:
            return True, UuidGenerator()
        if type_ == str:
            return False, StrSequenceGenerator()
        if type_ == int:
            return False, IntSequenceGenerator()
    elif type_ == datetime and name in ("created_at", "updated_at"):
        return False, TimestampGenerator()
    return True, None


def get_default_generator_for_update(
    name: str, type_: Type
) -> Tuple[bool, Optional[AttrValueGeneratorABC]]:
    if type_ == datetime and name == "updated_at":
        return False, TimestampGenerator()
    return True, None



import logging

from marshy.factory.impl_marshaller_factory import register_impl
from marshy.marshaller.obj_marshaller import AttrConfig
from marshy.marshaller.property_marshaller import PropertyConfig
from marshy.marshaller_context import MarshallerContext
from marshy_config_servey import raise_non_ignored

from servey.cache_control.cache_control_abc import CacheControlABC
from servey.cache_control.secure_hash_cache_control import SecureHashCacheControl
from servey.cache_control.timestamp_cache_control import TimestampCacheControl
from servey.cache_control.ttl_cache_control import TtlCacheControl

from persisty.attr.attr_filter import AttrFilter
from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC
from persisty.attr.generator.default_value_generator import DefaultValueGenerator
from persisty.attr.generator.timestamp_generator import TimestampGenerator
from persisty.attr.generator.uuid_generator import UuidGenerator
from persisty.finder.module_store_finder import ModuleStoreFinder
from persisty.finder.store_finder_abc import StoreFactoryFinderABC
from persisty.index.attr_index import AttrIndex
from persisty.index.index_abc import IndexABC
from persisty.index.unique_index import UniqueIndex
from persisty.key_config.attr_key_config import AttrKeyConfig
from persisty.key_config.composite_key_config import CompositeKeyConfig
from persisty.key_config.key_config_abc import KeyConfigABC
from persisty.link.belongs_to import BelongsTo
from persisty.link.has_count import HasCount
from persisty.link.has_many import HasMany
from persisty.link.link_abc import LinkABC
from persisty.search_filter.and_filter import And
from persisty.search_filter.exclude_all import ExcludeAll
from persisty.search_filter.include_all import IncludeAll
from persisty.search_filter.not_filter import Not
from persisty.search_filter.or_filter import Or
from persisty.search_filter.query_filter import QueryFilter
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.util import UNDEFINED

priority = 100
LOGGER = logging.getLogger(__name__)


def configure(context: MarshallerContext):
    # A bit hacky, but we want NONE in output, but NOT undefined
    AttrConfig.filter_dumped_values = (UNDEFINED,)
    PropertyConfig.filter_dumped_values = (UNDEFINED,)

    configure_search_filters(context)
    configure_key_configs(context)
    configure_attr_value_generator(context)
    configure_cache_control(context)
    configure_indexes(context)
    configure_links(context)
    configure_finders(context)

    configure_sqlalchemy(context)
    configure_celery(context)
    configure_serverless(context)


def configure_search_filters(context: MarshallerContext):
    register_impl(SearchFilterABC, And, context)
    register_impl(SearchFilterABC, Or, context)
    register_impl(SearchFilterABC, ExcludeAll, context)
    register_impl(SearchFilterABC, AttrFilter, context)
    register_impl(SearchFilterABC, IncludeAll, context)
    register_impl(SearchFilterABC, Not, context)
    register_impl(SearchFilterABC, QueryFilter, context)


def configure_key_configs(context: MarshallerContext):
    register_impl(KeyConfigABC, CompositeKeyConfig, context)
    register_impl(KeyConfigABC, AttrKeyConfig, context)


def configure_attr_value_generator(context: MarshallerContext):
    register_impl(AttrValueGeneratorABC, DefaultValueGenerator, context)
    register_impl(AttrValueGeneratorABC, TimestampGenerator, context)
    register_impl(AttrValueGeneratorABC, UuidGenerator, context)
    register_impl(AttrValueGeneratorABC, TimestampGenerator, context)
    register_impl(AttrValueGeneratorABC, UuidGenerator, context)


def configure_cache_control(context: MarshallerContext):
    register_impl(CacheControlABC, SecureHashCacheControl, context)
    register_impl(CacheControlABC, TimestampCacheControl, context)
    register_impl(CacheControlABC, TtlCacheControl, context)


def configure_indexes(context: MarshallerContext):
    register_impl(IndexABC, AttrIndex, context)
    register_impl(IndexABC, UniqueIndex, context)
    try:
        from persisty.impl.dynamodb.partition_sort_index import PartitionSortIndex

        register_impl(IndexABC, PartitionSortIndex, context)
    except ImportError as e:
        raise_non_ignored(e)


def configure_links(context: MarshallerContext):
    register_impl(LinkABC, BelongsTo, context)
    register_impl(LinkABC, HasMany, context)
    register_impl(LinkABC, HasCount, context)


def configure_sqlalchemy(context: MarshallerContext):
    try:
        # Local import in case sqlalchemy is not included (Optional extra)
        from marshy_config_persisty import sqlalchemy_config

        sqlalchemy_config.configure_converters(context)
        sqlalchemy_config.configure_sqlalchemy_context(context)
    except ImportError as e:
        msg = str(e)
        if msg.startswith("No module named '"):
            module_name = msg[len("No module named '") : -1]
            if module_name == "sqlalchemy":
                return
        raise_non_ignored(e)


def configure_finders(context: MarshallerContext):
    register_impl(StoreFactoryFinderABC, ModuleStoreFinder, context)


def configure_celery(context: MarshallerContext):
    try:
        from servey.servey_celery.celery_config.celery_config_abc import CeleryConfigABC
        from persisty.trigger.celery_store_trigger_config import (
            CeleryStoreTriggerConfig,
        )

        register_impl(CeleryConfigABC, CeleryStoreTriggerConfig, context)
    except ImportError as e:
        raise_non_ignored(e)


def configure_serverless(context: MarshallerContext):
    try:
        from servey.servey_aws.serverless.yml_config.yml_config_abc import YmlConfigABC
        from persisty.migration.serverless.dynamodb_yml_config import DynamodbYmlConfig

        register_impl(YmlConfigABC, DynamodbYmlConfig, context)
    except ImportError as e:
        raise_non_ignored(e)



import dataclasses
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Tuple, Type, Iterator, TypeVar, Dict

from marshy.factory.dataclass_marshaller_factory import dataclass_marshaller
from marshy.marshaller_context import MarshallerContext
from schemey import SchemaContext, Schema
from servey.cache_control.cache_control_abc import CacheControlABC
from servey.cache_control.secure_hash_cache_control import SecureHashCacheControl

from persisty.attr.attr import Attr
from persisty.index.index_abc import IndexABC
from persisty.key_config.attr_key_config import ATTR_KEY_CONFIG
from persisty.key_config.key_config_abc import KeyConfigABC
from persisty.link.link_abc import LinkABC
from persisty.store_access import StoreAccess, ALL_ACCESS

from persisty.util import to_camel_case
from persisty.util.undefined import UNDEFINED

T = TypeVar("T")


# pylint: disable=R0902
@dataclass
class StoreMeta:
    """
    Metadata object for a store. Contains info on the type of data being stored.
    """

    name: str
    attrs: Tuple[Attr, ...]
    key_config: KeyConfigABC = ATTR_KEY_CONFIG
    store_access: StoreAccess = ALL_ACCESS
    cache_control: CacheControlABC = SecureHashCacheControl()
    batch_size: int = 100
    description: Optional[str] = None
    links: Tuple[LinkABC, ...] = tuple()
    indexes: Tuple[IndexABC, ...] = tuple()

    def get_stored_dataclass(self) -> Type:
        return self._get_dataclass(
            "_stored_dataclass",
            self.name.title().replace("_", ""),
            iter(self.attrs),
            self.links,
        )

    def get_read_dataclass(self) -> Type:
        attrs = (a for a in self.attrs if a.readable)
        return self._get_dataclass(
            "_read_dataclass", self.name.title().replace("_", ""), attrs, self.links
        )

    def get_create_dataclass(self) -> Type:
        attrs = (a for a in self.attrs if a.creatable)
        return self._get_dataclass(
            "_create_dataclass",
            self.name.title().replace("_", "") + "Create",
            attrs,
        )

    def get_update_dataclass(self) -> Type:
        attrs = (a for a in self.attrs if a.updatable)
        return self._get_dataclass(
            "_update_dataclass",
            self.name.title().replace("_", "") + "Update",
            attrs,
        )

    def get_search_filter_factory_dataclass(self) -> Optional[Type]:
        search_filter_factory_dataclass = getattr(
            self, "_search_filter_factory_dataclass", UNDEFINED
        )
        if search_filter_factory_dataclass is UNDEFINED:
            from persisty.search_filter.search_filter_factory import (
                search_filter_dataclass_for,
            )

            search_filter_factory_dataclass = search_filter_dataclass_for(self)
            setattr(
                self,
                "_search_filter_factory_dataclass",
                search_filter_factory_dataclass,
            )
        return search_filter_factory_dataclass

    def get_sort_order_factory_dataclass(self) -> Optional[Type]:
        search_order_factory_dataclass = getattr(
            self, "_search_order_factory_dataclass", UNDEFINED
        )
        if search_order_factory_dataclass is UNDEFINED:
            from persisty.search_order.search_order_factory import (
                search_order_dataclass_for,
            )

            search_order_factory_dataclass = search_order_dataclass_for(self)
            setattr(
                self, "_search_order_factory_dataclass", search_order_factory_dataclass
            )
        return search_order_factory_dataclass

    def get_sortable_attrs_as_enum(self) -> Optional[Enum]:
        result = getattr(self, "_sortable_attrs", UNDEFINED)
        if result is UNDEFINED:
            attrs = {f.name: f.name for f in self.attrs if f.sortable}
            if attrs:
                result = Enum(f"{to_camel_case(self.name)}Sortable", attrs)
            else:
                result = None
            setattr(self, "_sortable_attrs", result)
        return result

    def _get_dataclass(
        self,
        attr_name: str,
        name: str,
        attrs: Iterator[Attr],
        links: Optional[Tuple[LinkABC, ...]] = None,
    ) -> Optional[Type]:
        result = getattr(self, attr_name, None)
        if result is None:
            annotations = {}
            params = {"__annotations__": annotations, "__stored_type__": attr_name}
            for attr in attrs:
                params[attr.name] = attr.to_field()
                annotations[attr.name] = attr.schema.python_type
            if annotations:
                if links:
                    for link in links:
                        params[link.get_name()] = link
                if self.description:
                    params["__doc__"] = self.description
                params["__persisty_store_meta__"] = self
                params["__schema_factory__"] = _schema_factory
                params["__marshaller_factory__"] = _marshaller_factory
                params["__eq__"] = _eq
                result = dataclass(type(name, tuple(), params))
            else:
                result = None
            setattr(self, attr_name, result)
        return result


def get_meta(type_: Type) -> Optional[StoreMeta]:
    meta = getattr(type_, "__persisty_store_meta__", None)
    return meta


def get_stored_dataclass(type_: Type[T]) -> Type[T]:
    # noinspection PyTypeChecker
    return get_meta(type_).get_stored_dataclass()


def get_read_dataclass(type_: Type[T]) -> Type[T]:
    # noinspection PyTypeChecker
    return get_meta(type_).get_read_dataclass()


def get_create_dataclass(type_: Type[T]) -> Type[T]:
    # noinspection PyTypeChecker
    return get_meta(type_).get_create_dataclass()


def get_update_dataclass(type_: Type[T]) -> Type[T]:
    # noinspection PyTypeChecker
    return get_meta(type_).get_update_dataclass()


# pylint: disable=W0613
# noinspection PyDecorator,PyUnusedLocal
@classmethod
def _schema_factory(
    cls, context: SchemaContext, path: str, ref_schemas: Dict[Type, Schema]
):
    """Override the default schemey behavior and do not include defaults with these generated dataclasses"""
    # noinspection PyDataclass
    schema = {
        "type": "object",
        "properties": {
            f.name: f.metadata.get("schemey").schema for f in dataclasses.fields(cls)
        },
        "additionalProperties": False,
    }
    if cls.__doc__:
        schema["description"] = cls.__doc__.strip()
    schema = Schema(schema, cls)
    return schema


# noinspection PyDecorator
@classmethod
def _marshaller_factory(cls, marshaller_context: MarshallerContext):
    return dataclass_marshaller(
        cls, marshaller_context, exclude_dumped_values=(UNDEFINED,)
    )


def _eq(self, other):
    meta: StoreMeta = get_meta(self)
    for attr in meta.attrs:
        a = getattr(self, attr.name, UNDEFINED)
        b = getattr(other, attr.name, UNDEFINED)
        if a != b:
            return False
    return True



import operator
from enum import Enum
from functools import partial

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshaller_context import MarshallerContext

from persisty.util.undefined import UNDEFINED
from persisty.attr.attr_type import AttrType


def _gte(a, b):
    return a >= b


def _lte(a, b):
    return a <= b


def _contains(a, b):
    return str(b) in str(a)


def _startswith(a, b):
    return str(a).startswith(str(b))


def _endswith(a, b):
    return str(a).endswith(str(b))


def _oneof(a, b):
    return a in b


def _exists(a, _):
    return a not in (None, UNDEFINED)


def _not_exists(a, _):
    return a in (None, UNDEFINED)


class AttrFilterOp(Enum):
    """
    Operations which attr search_filter supports. Provides a definitive set of limited attributes that storage
    implementations may implement. (e.g.: sql, dynamodb)
    """

    contains = partial(_contains)
    endswith = partial(_endswith)
    eq = operator.eq
    exists = partial(_exists)
    gt = operator.gt
    gte = partial(_gte)
    lt = operator.lt
    lte = partial(_lte)
    ne = operator.ne
    not_exists = partial(_not_exists)
    oneof = partial(_oneof)
    startswith = partial(_startswith)

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    @classmethod
    def __marshaller_factory__(cls, marshaller_context: MarshallerContext):
        return _AttrFilterOpMarshaller()


FILTER_OPS = (
    AttrFilterOp.eq,
    AttrFilterOp.exists,
    AttrFilterOp.ne,
    AttrFilterOp.not_exists,
)
SORTABLE_FILTER_OPS = FILTER_OPS + (
    AttrFilterOp.gt,
    AttrFilterOp.gte,
    AttrFilterOp.lt,
    AttrFilterOp.lte,
)
STRING_FILTER_OPS = SORTABLE_FILTER_OPS + (
    AttrFilterOp.startswith,
    AttrFilterOp.endswith,
    AttrFilterOp.contains,
)
TYPE_FILTER_OPS = {
    AttrType.BOOL: FILTER_OPS,
    AttrType.DATETIME: SORTABLE_FILTER_OPS,
    AttrType.FLOAT: SORTABLE_FILTER_OPS,
    AttrType.INT: SORTABLE_FILTER_OPS,
    AttrType.JSON: FILTER_OPS,
    AttrType.STR: STRING_FILTER_OPS,
    AttrType.UUID: FILTER_OPS,
}
NONE_TYPE = type(None)

SORTABLE_TYPES = frozenset(
    (
        AttrType.BOOL,
        AttrType.DATETIME,
        AttrType.FLOAT,
        AttrType.INT,
        AttrType.STR,
    )
)


class _AttrFilterOpMarshaller(MarshallerABC[AttrFilterOp]):
    def __init__(self):
        super().__init__(AttrFilterOp)

    def load(self, item: str) -> AttrFilterOp:
        return AttrFilterOp[item]

    def dump(self, item: AttrFilterOp) -> str:
        return item.name



from dataclasses import dataclass
from typing import Dict, Iterator

from sqlalchemy import ForeignKeyConstraint

from persisty.key_config.attr_key_config import AttrKeyConfig
from persisty.link.belongs_to import BelongsTo
from persisty.link.on_delete import OnDelete
from persisty.store_meta import StoreMeta


@dataclass
class SqlalchemyConstraintConverter:
    schema: Dict[str, StoreMeta]

    def get_foreign_key_constraints(
        self, store_meta: StoreMeta
    ) -> Iterator[ForeignKeyConstraint]:
        for link in store_meta.links:
            if isinstance(link, BelongsTo):
                linked_store = self.schema.get(link.linked_store_name)
                if linked_store and isinstance(linked_store.key_config, AttrKeyConfig):
                    kwargs = {}
                    if link.on_delete == OnDelete.CASCADE:
                        kwargs["ondelete"] = "CASCADE"
                    elif link.on_delete == OnDelete.NULLIFY:
                        kwargs["ondelete"] = "SET NULL"
                    constraint = ForeignKeyConstraint(
                        (link.key_attr_name,),
                        (
                            f"{link.linked_store_name}.{linked_store.key_config.attr_name}",
                        ),
                        **kwargs,
                    )
                    yield constraint



from abc import ABC, abstractmethod
from typing import Optional

from marshy.factory.impl_marshaller_factory import get_impls

from persisty.impl.sqlalchemy.sqlalchemy_context import SqlalchemyContext

_DEFAULT_CONTEXT = None


class SqlalchemyContextFactoryABC(ABC):
    priority: int = 100

    @abstractmethod
    def create(self) -> Optional[SqlalchemyContext]:
        """Create a new context"""


def create_default_context() -> Optional[SqlalchemyContext]:
    factories = list(get_impls(SqlalchemyContextFactoryABC))
    factories.sort(key=lambda f: f.priority, reverse=True)
    for factory in factories:
        context = factory().create()
        if context:
            return context


# pylint: disable=W0603
def get_default_context() -> Optional[SqlalchemyContext]:
    global _DEFAULT_CONTEXT
    if not _DEFAULT_CONTEXT:
        _DEFAULT_CONTEXT = create_default_context()
    return _DEFAULT_CONTEXT



from logging import Logger
from typing import Optional, FrozenSet

from persisty.store.store_abc import StoreABC
from persisty.util.logify import logify

METHODS: FrozenSet[str] = frozenset(
    name for name in StoreABC.__dict__ if name != "get_meta"
)


def logging_store(store: StoreABC, logger: Optional[Logger] = None) -> StoreABC:
    return logify(store, METHODS, logger)



import os
from copy import deepcopy
from decimal import Decimal
from typing import Optional, Dict, Tuple, List, Set
from dataclasses import dataclass, field

import boto3
import marshy
from boto3.dynamodb.conditions import Not as DynNot, ConditionBase, Key
from botocore.exceptions import ClientError
from marshy.types import ExternalItemType

from persisty.attr.attr import Attr
from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC
from persisty.errors import PersistyError
from persisty.impl.dynamodb.partition_sort_index import PartitionSortIndex
from persisty.search_filter.and_filter import And
from persisty.search_filter.exclude_all import EXCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.attr.attr_filter import AttrFilter, AttrFilterOp
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_order.search_order import SearchOrder
from persisty.store.store_abc import StoreABC, T
from persisty.store_meta import StoreMeta
from persisty.util import filter_none, get_logger
from persisty.util.undefined import UNDEFINED

logger = get_logger(__name__)


def catch_client_error(fn):
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ClientError as e:
            raise PersistyError(e) from e

    return wrapper


# pylint: disable=R0902
@dataclass(frozen=True)
class DynamodbTableStore(StoreABC[T]):
    """Store backed by a dynamodb table. Does not do single table design or anything of that nature."""

    meta: StoreMeta
    table_name: str
    index: PartitionSortIndex
    global_secondary_indexes: Dict[str, PartitionSortIndex] = field(
        default_factory=dict
    )
    aws_profile_name: Optional[str] = None
    region_name: Optional[str] = field(
        default_factory=lambda: os.environ.get("AWS_REGION")
    )
    decimal_format: str = "%.9f"
    max_local_search_size: int = None

    def __post_init__(self):
        if self.max_local_search_size is None:
            object.__setattr__(self, "max_local_search_size", self.meta.batch_size * 5)

    def get_meta(self) -> StoreMeta:
        return self.meta

    @catch_client_error
    def create(self, item: T) -> T:
        item = self._dump_create(item)
        self._dynamodb_table().put_item(
            Item=item,
            ConditionExpression=DynNot(self.index.to_condition_expression(item)),
        )
        loaded = self._load(item)
        return loaded

    @catch_client_error
    def read(self, key: str) -> Optional[T]:
        table = self._dynamodb_table()
        if not isinstance(key, str):
            key = str(key)
        key_dict = self.meta.key_config.to_key_dict(key)
        response = table.get_item(Key=key_dict)
        loaded = self._load(response.get("Item"))
        return loaded

    @catch_client_error
    def read_batch(self, keys: List[str]) -> List[Optional[T]]:
        assert len(keys) <= self.meta.batch_size
        key_config = self.meta.key_config
        resource = self._dynamodb_resource()

        kwargs = {
            "RequestItems": {
                self.table_name: {
                    "Keys": [key_config.to_key_dict(key) for key in set(keys)]
                }
            }
        }
        results_by_key = {}
        response = resource.batch_get_item(**kwargs)
        for item in response["Responses"][self.table_name]:
            loaded = self._load(item)
            key = key_config.to_key_str(loaded)
            results_by_key[key] = loaded
        assert not response.get(
            "UnprocessedKeys"
        )  # Batch size would have been greater than 16 Mb
        return [results_by_key.get(key) for key in keys]

    @catch_client_error
    def update(
        self, updates: T, precondition: SearchFilterABC = INCLUDE_ALL
    ) -> Optional[T]:
        if precondition is not INCLUDE_ALL:
            search_filter = precondition.lock_attrs(self.meta.attrs)
            item = self.read(self.meta.key_config.to_key_str(updates))
            if not search_filter.match(item, self.meta.attrs):
                return None
        updates = self._dump_update(updates)
        key_dict = self.index.to_dict(updates)
        table = self._dynamodb_table()
        updates = {**updates}
        updates.pop(self.index.pk)
        if self.index.sk:
            updates.pop(self.index.sk)
        update = _build_update(updates)
        response = table.update_item(
            Key=key_dict,
            ConditionExpression=self.index.to_condition_expression(key_dict),
            UpdateExpression=update["str"],
            ExpressionAttributeNames=update["names"],
            ExpressionAttributeValues=update["values"],
            ReturnValues="ALL_NEW",
        )
        loaded = self._load(response.get("Attributes"))
        return loaded

    def _update(
        self,
        key: str,
        item: T,
        updates: T,
        search_filter: SearchFilterABC = INCLUDE_ALL,
    ) -> Optional[T]:
        return self.update(updates, search_filter)

    @catch_client_error
    def delete(self, key: str) -> bool:
        table = self._dynamodb_table()
        key_dict = self.meta.key_config.to_key_dict(key)
        response = table.delete_item(Key=key_dict, ReturnValues="ALL_OLD")
        attributes = response.get("Attributes")
        return bool(attributes)

    def _delete(self, key: str, item: T) -> bool:
        return self.delete(key)

    @catch_client_error
    def search(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        if limit is None:
            limit = self.meta.batch_size
        assert limit <= self.meta.batch_size
        search_filter = search_filter.lock_attrs(self.meta.attrs)
        if search_order:
            search_order.validate_for_attrs(self.meta.attrs)
        if search_filter is EXCLUDE_ALL:
            return ResultSet([])
        index_name, index = self._get_index_for_search(search_filter, search_order)
        key_filter, other_filter = _separate_index_from_filter(index, search_filter)
        if other_filter:
            (
                filter_expression,
                search_filter_handled_natively,
            ) = other_filter.build_filter_expression(self.meta.attrs)
        else:
            filter_expression = None
            search_filter_handled_natively = True
        query_args = filter_none(
            {
                "KeyConditionExpression": self._to_key_condition_expression(key_filter),
                "IndexName": index_name,
                "Select": "SPECIFIC_ATTRIBUTES",
                "ProjectionExpression": ",".join(
                    a.name for a in self.meta.attrs if a.readable
                ),
                "FilterExpression": filter_expression,
                "ScanIndexForward": _get_scan_index_forward(index, search_order),
            }
        )
        search_order_handled_natively = _is_search_order_handled_natively(
            index, search_order
        )
        if search_order_handled_natively:
            return self._search_native_order(
                query_args,
                index,
                other_filter,
                search_filter_handled_natively,
                page_key,
                limit,
            )
        return self._search_local_order(
            query_args,
            index,
            other_filter,
            search_filter_handled_natively,
            search_order,
            page_key,
            limit,
        )

    # pylint: disable=R0913
    def _search_native_order(
        self,
        query_args: Dict,
        index: Optional[PartitionSortIndex],
        search_filter: SearchFilterABC,
        search_filter_handled_natively: bool,
        page_key: Optional[str],
        limit: int,
    ) -> ResultSet[T]:
        if page_key:
            query_args["ExclusiveStartKey"] = self.meta.key_config.to_key_dict(page_key)
        table = self._dynamodb_table()
        results = []
        while True:
            response = _get_search_response(table, index, query_args)
            items = self._load_items(
                response, search_filter, search_filter_handled_natively
            )
            results.extend(items)
            if len(results) >= limit:
                results = results[:limit]
                next_page_key = self.meta.key_config.to_key_str(results[-1])
                assert next_page_key != page_key  # Paranoid prevent infinite loop!
                return ResultSet(results, next_page_key)
            last_evaluated_key = response.get("LastEvaluatedKey")
            if not last_evaluated_key:
                return ResultSet(results)
            query_args["ExclusiveStartKey"] = last_evaluated_key

    # pylint: disable=R0914
    def _search_local_order(
        self,
        query_args: Dict,
        index: Optional[PartitionSortIndex],
        search_filter: SearchFilterABC,
        search_filter_handled_natively: bool,
        search_order: SearchOrder,
        page_key: Optional[str],
        limit: int,
    ) -> ResultSet[T]:
        table = self._dynamodb_table()
        results = []
        while True:
            response = _get_search_response(table, index, query_args)
            items = self._load_items(
                response, search_filter, search_filter_handled_natively
            )
            results.extend(items)
            if len(items) > self.max_local_search_size:
                raise PersistyError("sort_failed")
            last_evaluated_key = response.get("LastEvaluatedKey")
            if not last_evaluated_key:
                results = list(search_order.sort(results))
                key_config = self.meta.key_config
                offset = 0
                if page_key:
                    offset = next(
                        i + 1
                        for i, result in enumerate(results)
                        if key_config.to_key_str(result) == page_key
                    )
                next_page_key = None
                if len(results) > offset + limit:
                    next_page_key = key_config.to_key_str(results[offset + limit - 1])
                results = results[offset : (offset + limit)]
                return ResultSet(results, next_page_key)
            query_args["ExclusiveStartKey"] = last_evaluated_key

    @catch_client_error
    def count(self, search_filter: SearchFilterABC = INCLUDE_ALL) -> int:
        search_filter = search_filter.lock_attrs(self.meta.attrs)
        if search_filter is EXCLUDE_ALL:
            return 0
        index_name, index = self._get_index_for_search(search_filter, None)
        key_filter, other_filter = _separate_index_from_filter(index, search_filter)
        if other_filter:
            (
                filter_expression,
                search_filter_handled_natively,
            ) = other_filter.build_filter_expression(self.meta.attrs)
        else:
            filter_expression = None
            search_filter_handled_natively = True
        if not search_filter_handled_natively:
            result = sum(1 for _ in self.search_all(search_filter))
            return result
        kwargs = filter_none(
            {
                "KeyConditionExpression": self._to_key_condition_expression(key_filter),
                "IndexName": index_name,
                "Select": "COUNT",
                "FilterExpression": filter_expression,
            }
        )
        table = self._dynamodb_table()
        count = 0
        while True:
            if index:
                response = table.query(**kwargs)
            else:
                response = table.scan(**kwargs)
            count += response["Count"]  # Items
            last_evaluated_key = response.get("LastEvaluatedKey")
            kwargs["ExclusiveStartKey"] = last_evaluated_key
            if not last_evaluated_key:
                return count

    def _to_key_condition_expression(self, key_filter: Optional[AttrFilter]):
        if not key_filter:
            return
        attr = next(a for a in self.meta.attrs if a.name == key_filter.name)
        marshy.dump(key_filter.value, attr.schema.python_type)
        value = marshy.dump(key_filter.value, attr.schema.python_type)
        return Key(key_filter.name).eq(value)

    def _edit_batch(
        self, edits: List[BatchEdit], items_by_key: Dict[str, T]
    ) -> List[BatchEditResult]:
        assert len(edits) <= self.meta.batch_size
        results = []
        key_config = self.meta.key_config
        table = self._dynamodb_table()
        with table.batch_writer() as batch:
            for edit in edits:
                if edit.create_item:
                    item = self._dump_create(edit.create_item)
                    batch.put_item(Item=item)
                    results.append(BatchEditResult(edit, True))
                elif edit.update_item:
                    updates = edit.update_item
                    key = key_config.to_key_str(updates)
                    item = items_by_key[key]
                    to_put = {}
                    for attr in self.meta.attrs:
                        value = UNDEFINED
                        if attr.update_generator:
                            if attr.updatable:
                                value = attr.update_generator.transform(
                                    getattr(updates, attr.name), updates
                                )
                            else:
                                value = attr.update_generator.transform(
                                    UNDEFINED, updates
                                )
                        elif attr.updatable:
                            value = getattr(updates, attr.name)
                            if value is UNDEFINED:
                                value = getattr(item, attr.name)
                        if value is not UNDEFINED:
                            setattr(item, attr.name, value)
                            value = marshy.dump(value, attr.schema.python_type)
                            to_put[attr.name] = self._convert_to_decimals(value)
                    to_put = self._convert_to_decimals(to_put)
                    batch.put_item(Item=to_put)
                    edit.update_item = deepcopy(item)  # In case of multi put
                    results.append(BatchEditResult(edit, True))
                else:
                    key = key_config.to_key_dict(edit.delete_key)
                    batch.delete_item(Key=key)
                    results.append(BatchEditResult(edit, True))
        return results

    def _load(self, item) -> T:
        if item is None:
            return None
        item = self._convert_from_decimals(item)
        kwargs = {}
        for attr in self.meta.attrs:
            value = item.get(attr.name, UNDEFINED)
            if attr.readable and value is not UNDEFINED:
                # noinspection PyTypeChecker
                kwargs[attr.name] = marshy.load(attr.schema.python_type, value)
        result = self.meta.get_read_dataclass()(**kwargs)
        return result

    def _convert_from_decimals(self, item):
        if isinstance(item, dict):
            return {k: self._convert_from_decimals(v) for k, v in item.items()}
        if isinstance(item, list):
            return [self._convert_from_decimals(i) for i in item]
        if isinstance(item, Decimal):
            int_val = int(item)
            if int_val == item:
                return int_val
            return float(item)
        return item

    def _dump_create(self, to_create: T):
        result = {}
        for attr in self.meta.attrs:
            self._dump_attr(
                to_create, attr, attr.creatable, attr.create_generator, result
            )
        return result

    def _dump_update(self, to_update: T):
        result = {}
        for attr in self.meta.attrs:
            self._dump_attr(
                to_update, attr, attr.updatable, attr.update_generator, result
            )
        return result

    def _dump_attr(
        self,
        item: T,
        attr: Attr,
        accepts_input: bool,
        generator: Optional[AttrValueGeneratorABC],
        target: ExternalItemType,
    ):
        value = UNDEFINED
        if accepts_input:
            value = getattr(item, attr.name, UNDEFINED)
        if generator:
            value = generator.transform(value, item)
        if value is not UNDEFINED:
            value = marshy.dump(value, attr.schema.python_type)
            target[attr.name] = self._convert_to_decimals(value)

    def _convert_to_decimals(self, item):
        if isinstance(item, dict):
            return {k: self._convert_to_decimals(v) for k, v in item.items()}
        if isinstance(item, list):
            return [self._convert_to_decimals(i) for i in item]
        if isinstance(item, float):
            int_val = int(item)
            if int_val == item:
                return int_val
            return Decimal(self.decimal_format % item)
        return item

    def _get_index_for_search(
        self,
        search_filter: SearchFilterABC,
        search_order: Optional[SearchOrder],
    ) -> Tuple[Optional[str], Optional[PartitionSortIndex]]:
        eq_attr_names = _get_top_level_eq_attrs(search_filter)
        sort_attr_names = (
            {s.attr for s in search_order.orders} if search_order else set()
        )
        name = None
        score = _get_score_for_index(self.index, eq_attr_names, sort_attr_names)
        index = self.index if score else None
        for gsi_name, gsi in self.global_secondary_indexes.items():
            gsi_score = _get_score_for_index(gsi, eq_attr_names, sort_attr_names)
            if gsi_score > score:
                name = gsi_name
                score = gsi_score
                index = gsi
        if not score:
            index = None
        return name, index

    def _dynamodb_table(self):
        if hasattr(self, "_table"):
            return self._table
        resource = self._dynamodb_resource()
        table = resource.Table(self.table_name)
        object.__setattr__(self, "_table", table)
        return table

    def _dynamodb_resource(self):
        if hasattr(self, "_resource"):
            return self._resource
        kwargs = filter_none(
            {"profile_name": self.aws_profile_name, "region_name": self.region_name}
        )
        session = boto3.Session(**kwargs)
        resource = session.resource("dynamodb")
        object.__setattr__(self, "_resource", resource)
        return resource

    def _load_items(self, response, search_filter, search_filter_handled_natively):
        items = [self._load(item) for item in response["Items"]]
        if not search_filter_handled_natively:
            items = [
                item for item in items if search_filter.match(item, self.meta.attrs)
            ]
        return items


def _build_update(updates: dict):
    update_str = "set "
    update_list = []
    names = {}
    values = {}
    for k, v in updates.items():
        update_list.append(f"#n_{k} = :v_{k}")
        names[f"#n_{k}"] = k
        values[f":v_{k}"] = v
    update_str += ", ".join(update_list)
    return {"str": update_str, "names": names, "values": values}


def _get_top_level_eq_attrs(search_filter: SearchFilterABC) -> List[str]:
    if isinstance(search_filter, AttrFilter) and search_filter.op == AttrFilterOp.eq:
        return [search_filter.name]
    if isinstance(search_filter, And):
        return [
            f.name
            for f in search_filter.search_filters
            if isinstance(f, AttrFilter) and f.op == AttrFilterOp.eq
        ]
    return []


def _separate_index_filters(
    index: PartitionSortIndex, eq_filters: List[AttrFilter]
) -> Tuple[ConditionBase, SearchFilterABC, bool]:
    index_filters = [f for f in eq_filters if f.name == index.pk]
    value = index_filters[0].value
    if value.__class__ not in (str, int, float, bool):
        value = marshy.dump(value)
    condition = Key(index.pk).eq(value)
    non_index_filters = tuple(f for f in eq_filters if f.name != index.pk)
    non_index_filter = None
    if non_index_filters:
        non_index_filter = And(non_index_filters)
    handled = len(index_filters) == 1
    return condition, non_index_filter, handled


def _get_score_for_index(
    index: PartitionSortIndex, eq_attrs: Set[str], sort_attrs: Set[str]
):
    if index.pk not in eq_attrs:
        return 0
    if index.sk in sort_attrs:
        return 20
    return 10


def _separate_index_from_filter(
    index: Optional[PartitionSortIndex], search_filter: SearchFilterABC
) -> Tuple[Optional[AttrFilter], Optional[SearchFilterABC]]:
    if not index:
        return None, search_filter
    if isinstance(search_filter, AttrFilter):
        return search_filter, None
    index_filter = None
    filters = []
    # noinspection PyUnresolvedReferences
    for f in search_filter.search_filters:
        if f.name == index.pk:
            index_filter = f
        else:
            filters.append(f)
    filter_expression = And(filters) if filters else None
    return index_filter, filter_expression


def _get_scan_index_forward(
    index: Optional[PartitionSortIndex], search_order: Optional[SearchOrder]
) -> Optional[bool]:
    if search_order and _is_search_order_handled_natively(index, search_order):
        return not search_order.orders[0].desc


def _is_search_order_handled_natively(
    index: Optional[PartitionSortIndex], search_order: Optional[SearchOrder]
) -> bool:
    if not search_order:
        return True
    if len(search_order.orders) > 1 or not index:
        return False
    return search_order.orders[0].attr == index.sk


def _get_search_response(table, index, query_args):
    if index:
        response = table.query(**query_args)
    else:
        response = table.scan(**query_args)
    return response



from datetime import datetime
from enum import Enum
from uuid import UUID

from marshy.factory.optional_marshaller_factory import get_optional_type


class AttrType(Enum):
    BINARY = "binary"
    BOOL = "bool"
    DATETIME = "datetime"
    FLOAT = "float"
    INT = "int"
    JSON = "json"
    STR = "str"
    UUID = "uuid"


TYPE_MAP = {
    bool: AttrType.BOOL,
    bytes: AttrType.BINARY,
    datetime: AttrType.DATETIME,
    float: AttrType.FLOAT,
    int: AttrType.INT,
    str: AttrType.STR,
    UUID: AttrType.UUID,
}
ATTR_TYPE_MAP = {v: k for k, v in TYPE_MAP.items()}


def attr_type(type_) -> AttrType:
    type_ = get_optional_type(type_) or type_
    result = TYPE_MAP.get(type_)
    if result:
        return result
    try:
        if issubclass(type_, Enum):
            return AttrType.STR
    except TypeError:
        pass
    return AttrType.JSON



from dataclasses import dataclass
from typing import Tuple

from persisty.index.index_abc import IndexABC


@dataclass
class UniqueIndex(IndexABC):
    """Unique attribute index"""

    attr_names: Tuple[str, ...]


def unique_index(*args):
    return UniqueIndex(args)



from dataclasses import dataclass
from typing import Tuple, Optional, Any

import marshy

from persisty.attr.attr import Attr
from persisty.attr.attr_filter_op import AttrFilterOp
from persisty.errors import PersistyError
from persisty.search_filter.search_filter_abc import SearchFilterABC, T


@dataclass(frozen=True)
class AttrFilter(SearchFilterABC[T]):
    name: str
    op: AttrFilterOp
    value: T

    def lock_attrs(self, attrs: Tuple[Attr, ...]) -> SearchFilterABC:
        try:
            attr = next(attr for attr in attrs if attr.name == self.name)
            assert attr.readable and self.op in attr.permitted_filter_ops
            value = attr.sanitize_type(self.value)
            return AttrFilter(self.name, self.op, value)
        except (StopIteration, AssertionError) as exc:
            raise PersistyError("attr_filter_invalid_for_attrs") from exc

    def match(self, item: T, attrs: Tuple[Attr, ...] = None) -> bool:
        value = getattr(item, self.name)
        try:
            result = self.op.value(value, self.value)
            return result
        except TypeError:
            return False  # Comparison failed

    def build_filter_expression(
        self, attrs: Tuple[Attr, ...]
    ) -> Tuple[Optional[Any], bool]:
        value = self.value
        from boto3.dynamodb.conditions import Attr as DynAttr

        attr = DynAttr(self.name)
        if self.op.name in {"contains", "eq", "gt", "gte", "lt", "lte", "ne"}:
            if not isinstance(value, str):
                value = marshy.dump(value)
            condition = getattr(attr, self.op.name)(value)
            return condition, True
        if self.op == AttrFilterOp.startswith:
            if not isinstance(value, str):
                value = marshy.dump(value)
            return attr.begins_with(value), True
        if self.op == AttrFilterOp.exists:
            return attr.exists(), True
        if self.op == AttrFilterOp.not_exists:
            return attr.not_exists(), True
        return None, False



from dataclasses import dataclass
from typing import Tuple

from persisty.attr.attr import Attr
from persisty.search_filter.exclude_all import EXCLUDE_ALL
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC, T


@dataclass(frozen=True)
class Not(SearchFilterABC[T]):
    search_filter: SearchFilterABC

    # pylint: disable=W0221
    def __new__(cls, search_filter: SearchFilterABC):
        """Strip out direct nested Not"""
        if isinstance(search_filter, Not):
            return search_filter.search_filter
        if search_filter is EXCLUDE_ALL:
            return INCLUDE_ALL
        if search_filter is INCLUDE_ALL:
            return EXCLUDE_ALL
        return super(Not, cls).__new__(cls)

    def lock_attrs(self, attrs: Tuple[Attr, ...]) -> SearchFilterABC:
        result = Not(self.search_filter.lock_attrs(attrs))
        return result

    def match(self, item: T, attrs: Tuple[Attr, ...]) -> bool:
        return not self.search_filter.match(item, attrs)



import dataclasses
from dataclasses import dataclass
from typing import Optional, Iterator, List, Dict

from persisty.errors import PersistyError
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.store.store_abc import StoreABC, T
from persisty.store_access import StoreAccess, ALL_ACCESS
from persisty.store_meta import StoreMeta


@dataclass
class RestrictAccessStore(StoreABC[T]):
    store: StoreABC[T]
    store_access: StoreAccess

    def get_store(self) -> StoreABC:
        return self.store

    def get_meta(self) -> StoreMeta:
        store_meta = getattr(self, "_store_meta", None)
        if not store_meta:
            store_meta = self.store.get_meta()
            store_meta = dataclasses.replace(
                store_meta, store_access=store_meta.store_access and self.store_access
            )
            setattr(self, "_store_meta", store_meta)
        return store_meta

    def get_store_access(self) -> StoreAccess:
        store_access = getattr(self, "_store_access", None)
        if not store_access:
            store_meta = self.store.get_meta()
            store_access = store_meta.store_access & self.store_access
            setattr(self, "_store_access", store_access)
        return store_access

    def create(self, item: T) -> Optional[T]:
        if not self.get_store_access().creatable:
            raise PersistyError("unavailable_operation")
        return self.store.create(item)

    def read(self, key: str) -> Optional[T]:
        if not self.get_store_access().readable:
            raise PersistyError("unavailable_operation")
        return self.store.read(key)

    def update(
        self, updates: T, precondition: SearchFilterABC = INCLUDE_ALL
    ) -> Optional[T]:
        if not self.get_store_access().updatable:
            raise PersistyError("unavailable_operation")
        return self.store.update(updates, precondition)

    # pylint: disable=W0212
    def _update(self, key: str, item: T, updates: T) -> Optional[T]:
        if not self.get_store_access().updatable:
            raise PersistyError("unavailable_operation")
        return self.store._update(key, item, updates)

    def delete(self, key: str) -> bool:
        if not self.get_store_access().deletable:
            raise PersistyError("unavailable_operation")
        return self.store.delete(key)

    # pylint: disable=W0212
    def _delete(self, key: str, item: T) -> bool:
        if not self.get_store_access().deletable:
            raise PersistyError("unavailable_operation")
        return self.store._delete(key, item)

    def count(self, search_filter: SearchFilterABC = INCLUDE_ALL) -> int:
        if not self.get_store_access().searchable:
            raise PersistyError("unavailable_operation")
        return self.store.count(search_filter)

    def search(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        if not self.get_store_access().searchable:
            raise PersistyError("unavailable_operation")
        return self.store.search(search_filter, search_order, page_key, limit)

    def search_all(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
    ) -> Iterator[T]:
        if not self.get_store_access().searchable:
            raise PersistyError("unavailable_operation")
        return self.store.search_all(search_filter, search_order)

    def edit_batch(self, edits: List[BatchEdit]) -> List[BatchEditResult]:
        _check_edits(edits, self.get_store_access())
        return self.store.edit_batch(edits)

    def _edit_batch(
        self, edits: List[BatchEdit], items_by_key: Dict[str, T]
    ) -> List[BatchEditResult]:
        _check_edits(edits, self.get_store_access())
        return self.store._edit_batch(edits, items_by_key)


def restrict_access_store(store: StoreABC, store_access: StoreAccess) -> StoreABC:
    if store_access == ALL_ACCESS:
        return store
    return RestrictAccessStore(store, store_access)


def _check_edits(edits: List[BatchEdit], store_access: StoreAccess):
    for edit in edits:
        create_error = edit.create_item and not store_access.creatable
        update_error = edit.update_item and not store_access.updatable
        delete_error = edit.delete_key and not store_access.deletable
        if create_error or update_error or delete_error:
            raise PersistyError("unavailable_operation")



import logging
import os
from typing import Optional

from sqlalchemy import create_engine

from persisty.impl.sqlalchemy.sqlalchemy_context import SqlalchemyContext
from persisty.impl.sqlalchemy.sqlalchemy_context_factory_abc import (
    SqlalchemyContextFactoryABC,
)

LOGGER = logging.getLogger(__name__)


class SqlalchemyContextFactory(SqlalchemyContextFactoryABC):
    # pylint: disable=W1203
    def create(self) -> Optional[SqlalchemyContext]:
        sql_urn = os.environ.get("PERSISTY_SQL_URN")
        # If we have not defined an sql urn, then we are in developer mode, where tables are created on the fly
        developer_mode = not sql_urn or os.environ.get("PERSISTY_DEVELOPER_MODE") == "1"
        if not sql_urn:
            sql_urn = "sqlite+pysqlite:///:memory:"
        sql_echo = (os.environ.get("SQL_ECHO") or "").lower() not in ("false", "0")
        if not os.environ.get("PERSISTY_SQL_URN"):
            LOGGER.warning(f"PERSISTY_SQL_URN NOT SET: USING {sql_urn}")
        engine = create_engine(sql_urn, echo=sql_echo, future=True)
        return SqlalchemyContext(engine, developer_mode)



from dataclasses import dataclass
from typing import Tuple, Optional, Any

from persisty.attr.attr import Attr
from persisty.search_filter.and_filter import build_filter_conditions
from persisty.search_filter.exclude_all import EXCLUDE_ALL
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC, T


@dataclass(frozen=True)
class Or(SearchFilterABC[T]):
    search_filters: Tuple[SearchFilterABC, ...]

    # pylint: disable=W0221
    def __new__(cls, search_filters: Tuple[SearchFilterABC, ...]):
        """Strip out nested And logic"""
        if not search_filters:
            return EXCLUDE_ALL
        if len(search_filters) == 1:
            return search_filters[0]
        flatten = next(
            (
                True
                for f in search_filters
                if isinstance(f, Or) or f is INCLUDE_ALL or f is EXCLUDE_ALL
            ),
            False,
        )
        if flatten:
            existing = set()
            flattened = []
            for f in search_filters:
                if f in existing:
                    continue
                existing.add(f)
                if isinstance(f, Or):
                    flattened.extend(f.search_filters)
                elif f is EXCLUDE_ALL:
                    continue
                elif f is INCLUDE_ALL:
                    return INCLUDE_ALL
                else:
                    flattened.append(f)
            if len(flattened) == 1:
                return flattened[0]
            return Or(tuple(flattened))
        return super(Or, cls).__new__(cls)

    def lock_attrs(self, attrs: Tuple[Attr, ...]) -> SearchFilterABC:
        result = Or(tuple(f.lock_attrs(attrs) for f in self.search_filters))
        return result

    def match(self, item: T, attrs: Tuple[Attr, ...]) -> bool:
        match = next(
            (True for f in self.search_filters if not f.match(item, attrs)), False
        )
        return match

    def build_filter_expression(
        self, attrs: Tuple[Attr, ...]
    ) -> Tuple[Optional[Any], bool]:
        conditions, all_handled = build_filter_conditions(self.search_filters, attrs)
        if not conditions:
            return None, False
        if len(conditions) == 1:
            condition = conditions[0]
        else:
            from boto3.dynamodb.conditions import Or as DynOr

            condition = DynOr(*conditions)
        return condition, all_handled



from typing import List, TypeVar, Generic, Type, Optional

from dataclasses import dataclass

T = TypeVar("T")


@dataclass
class ResultSet(Generic[T]):
    results: List[T]
    next_page_key: str = None


def result_set_dataclass_for(type_: Type[T]) -> Type[ResultSet[T]]:
    params = {
        "__annotations__": {
            "results": List[type_],
            "next_page_key": Optional[str],
        },
        "__doc__": f"Result Set of {type_.__name__}",
        "next_page_key": None,
    }
    type_name = f"{type_.__name__}ResultSet"
    type_ = dataclass(type(type_name, (ResultSet,), params))
    return type_



from typing import Optional, Dict, Iterator

from dataclasses import dataclass, field

from persisty.errors import PersistyError
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.store.store_abc import StoreABC, T
from persisty.store_meta import StoreMeta

from persisty.util.undefined import UNDEFINED


@dataclass
class MemStore(StoreABC[T]):
    """
    In memory store, used mostly for mocking, testing and caches. Relies on wrappers to provide access control,
    triggers and schema validation.
    """

    meta: StoreMeta = field()
    items: Dict[str, T] = field(default_factory=dict)

    def get_meta(self) -> StoreMeta:
        return self.meta

    def create(self, item: T) -> T:
        meta = self.meta

        kwargs = {}
        for attr in meta.attrs:
            value = UNDEFINED
            if attr.creatable:
                value = getattr(item, attr.name, UNDEFINED)
            if attr.create_generator:
                value = attr.create_generator.transform(value, item)
            if value is not UNDEFINED:
                value = attr.sanitize_type(value)
                kwargs[attr.name] = value
        stored_item = meta.get_stored_dataclass()(**kwargs)
        key = self.meta.key_config.to_key_str(stored_item)
        if key in (None, UNDEFINED):
            raise PersistyError(f"missing_key:{item}")
        if key in self.items:
            raise PersistyError(f"existing_value:{item}")
        self.items[key] = stored_item
        return self._load(stored_item)

    def read(self, key: str) -> Optional[T]:
        key = str(key)
        item = self.items.get(key)
        if item:
            item = self._load(item)
        return item

    def _update(
        self,
        key: str,
        item: T,
        updates: T,
    ) -> Optional[T]:
        stored_item = self.items.get(key)
        if stored_item:
            for attr in self.meta.attrs:
                value = UNDEFINED
                if attr.updatable:
                    value = getattr(updates, attr.name, UNDEFINED)
                if attr.update_generator:
                    value = attr.update_generator.transform(value, item)
                if value is not UNDEFINED:
                    value = attr.sanitize_type(value)
                    setattr(stored_item, attr.name, value)
            return self._load(stored_item)

    def _delete(self, key: str, item: T) -> bool:
        if key not in self.items:
            return False
        result = self.items.pop(key, UNDEFINED)
        return result is not UNDEFINED

    def search_all(
        self,
        search_filter: SearchFilterABC[T] = INCLUDE_ALL,
        search_order: Optional[SearchOrder[T]] = None,
    ) -> Iterator[T]:
        search_filter = search_filter.lock_attrs(self.meta.attrs)
        if search_order:
            search_order.validate_for_attrs(self.meta.attrs)
        items = self.items.values()
        if search_filter is not INCLUDE_ALL:
            items = (
                item for item in items if search_filter.match(item, self.meta.attrs)
            )
        items = [self._load(item) for item in items]
        if search_order and search_order.orders:
            items = search_order.sort(items)
        return iter(items)

    def count(self, search_filter: SearchFilterABC[T] = INCLUDE_ALL) -> int:
        search_filter = search_filter.lock_attrs(self.meta.attrs)
        if search_filter is INCLUDE_ALL:
            return len(self.items)
        count = sum(1 for _ in self.search_all(search_filter))
        return count

    def _load(self, item: T) -> T:
        kwargs = {}
        for attr in self.meta.attrs:
            if attr.readable:
                kwargs[attr.name] = getattr(item, attr.name)
        result = self.meta.get_read_dataclass()(**kwargs)
        return result



from dataclasses import dataclass
from typing import Any

from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC
from persisty.util.undefined import UNDEFINED


@dataclass
class DefaultValueGenerator(AttrValueGeneratorABC):
    default_value: Any

    def transform(self, value, item):
        return self.default_value if value is UNDEFINED else value



from abc import ABC, abstractmethod
from typing import Iterator, List, Optional, Dict

from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.search_order.search_order import SearchOrder
from persisty.store.store_abc import StoreABC, T
from persisty.stored import StoreMeta
from persisty.util import get_logger

logger = get_logger(__name__)


class WrapperStoreABC(StoreABC[T], ABC):
    @abstractmethod
    def get_store(self) -> StoreABC:
        """Get wrapped store"""

    def get_meta(self) -> StoreMeta:
        return self.get_store().get_meta()

    def create(self, item: T) -> T:
        return self.get_store().create(item)

    def read(self, key: str) -> Optional[T]:
        return self.get_store().read(key)

    def read_batch(self, keys: List[str]) -> List[Optional[T]]:
        return self.get_store().read_batch(keys)

    # pylint: disable=W0212
    def _update(self, key: str, item: T, updates: T) -> Optional[T]:
        return self.get_store()._update(key, item, updates)

    # pylint: disable=W0212
    def _delete(self, key: str, item: T) -> bool:
        return self.get_store()._delete(key, item)

    def search(
        self,
        search_filter: SearchFilterABC[T] = INCLUDE_ALL,
        search_order: Optional[SearchOrder[T]] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        store = self.get_store()
        return store.search(search_filter, search_order, page_key, limit)

    def search_all(
        self,
        search_filter: SearchFilterABC[T] = INCLUDE_ALL,
        search_order: Optional[SearchOrder[T]] = None,
    ) -> Iterator[T]:
        return self.get_store().search_all(search_filter, search_order)

    def count(self, search_filter: SearchFilterABC[T] = INCLUDE_ALL) -> int:
        return self.get_store().count(search_filter)

    # pylint: disable=W0212
    def _edit_batch(
        self, edits: List[BatchEdit[T, T]], items_by_key: Dict[str, T]
    ) -> List[BatchEditResult[T, T]]:
        return self.get_store()._edit_batch(edits, items_by_key)

    def edit_all(
        self, edits: Iterator[BatchEdit[T, T]]
    ) -> Iterator[BatchEditResult[T, T]]:
        return self.get_store().edit_all(edits)



"""
Use a key from an environment variable to do encryption at rest...
"""

import base64
import json
import os
import logging
from typing import Any

import pyaes


KEY_PARAM = "PERSISTY_SECRET_KEY"
logger = logging.getLogger(__name__)
_KEY = os.environ.get(KEY_PARAM)
if not _KEY:
    logging.getLogger().warning(
        f"NO SECRET KEY FOUND IN ENVIRONMENT SO USING DEFAULT. PLEASE SET {KEY_PARAM}! "
        f"THIS IS NOT SECURE!!!"
    )
    _KEY = "NOT_A_SECURE_KEY"
_KEY = _KEY.encode("utf-8")


def encrypt(item: Any) -> str:
    json_bytes = json.dumps(item).encode("utf-8")
    aes = pyaes.AESModeOfOperationCTR(_KEY)
    encrypted_bytes = aes.encrypt(json_bytes)
    b64_bytes = base64.b64encode(encrypted_bytes)
    b64_str = b64_bytes.decode("utf-8")
    return b64_str


def decrypt(encrypted: str) -> Any:
    b64_bytes = encrypted.encode("utf-8")
    encrypted_bytes = base64.b64decode(b64_bytes)
    aes = pyaes.AESModeOfOperationCTR(_KEY)
    json_bytes = aes.decrypt(encrypted_bytes)
    json_str = json_bytes.decode("utf-8")
    item = json.loads(json_str)
    return item



from dataclasses import dataclass, field
from typing import Optional, Type, Generic, TypeVar
from uuid import UUID, uuid4

from persisty.key_config.key_config_abc import KeyConfigABC

C = TypeVar("C")
U = TypeVar("U")


@dataclass
class BatchEdit(Generic[C, U]):
    """
    Batch edit should define one of create_item, update_item, or delete_key. Defining multiple is not valid.
    """

    id: UUID = field(default_factory=uuid4)
    create_item: Optional[C] = None
    update_item: Optional[U] = None
    delete_key: Optional[str] = None

    def get_key(self, key_config: KeyConfigABC) -> str:
        if self.create_item:
            return key_config.to_key_str(self.create_item)
        if self.update_item:
            return key_config.to_key_str(self.update_item)
        return self.delete_key


def batch_edit_dataclass_for(
    type_name: str, create_input_type: Type, update_input_type: Type
) -> Type:
    params = {
        "__annotations__": {
            "create_item": Optional[create_input_type],
            "update_item": Optional[update_input_type],
            "delete_key": Optional[str],
        },
        "__doc__": f"Batch edit of {type_name}",
        "create_item": None,
        "update_item": None,
        "delete_key": None,
    }
    type_ = dataclass(type(type_name, (), params))
    return type_
